(* errormsg.sml *)

signature ERRORMSG =
sig
    (* global variables of the error reporting mechanism *)
    val anyErrors : bool ref
    val fileName : string ref
    val lineNum : int ref
    val linePos : int list ref
    val sourceStream : TextIO.instream ref

    exception Error
      (* exception raised by fatalError *)
    val error : int -> string -> unit
      (* error pos message: print message string preceded by position
       * info in (line:column) format derived from character position pos *)
    val fatalError : int -> string -> 'a
      (* prints error message same as error, then aborts by raising Error *)
    val impossible : string -> 'a
      (* prints message, raises Error.  Intended to be called on paths
       * not expected to be taken *)
    val reset : string -> unit
      (* reset state of error machinery; takes file name as parameter *)
end

structure ErrorMsg : ERRORMSG =
struct

(* these are global reference variables that define the state of the
 * error reporting machinery *)

  val anyErrors = ref false
  val fileName = ref ""
  val lineNum = ref 1     (* number of last line *)
  val linePos = ref [1]   (* list of initial char positions of lines, in reverse order *)
  val sourceStream = ref TextIO.stdIn  (* default input source -- not used *)

  fun reset(fname: string) =
      (anyErrors:=false;
       fileName:=fname;
       lineNum:=1;
       linePos:=[1];
       sourceStream:=TextIO.stdIn)

  (* exception raised by fatalError and impossible *)
  exception Error

  (* converts character position pos into line number and column number
   * and prints them in format "<line>:<column>" *)
  fun printPos pos =
      let fun look(a::rest,n) =
		if a<pos then app print [":",
				       Int.toString n,
				       ".",
				       Int.toString (pos-a)]
		       else look(rest,n-1)
	    | look _ = print "0.0"
       in look(!linePos, !lineNum)
      end

  (* error: int -> string -> unit
   *   error pos msg:
   *      pos is a character position in the input
   *      msg is the error message to be printed with position info.
   *   The position pos is translated into line and column numbers using
   *   the line mapping represented by linePos and lineNum.
   *)
  fun error pos (msg:string) =
      (anyErrors := true;
       print (!fileName);
       printPos pos;
       print ":"; print msg; print "\n")

  fun fatalError pos (msg:string) =
      (error pos msg;
       raise Error)

  fun impossible msg =
      (app print ["Error: Compiler bug: ",msg,"\n"];
       TextIO.flushOut TextIO.stdOut;
       raise Error)

end  (* structure ErrorMsg *)
  
